@@ -277,6 +277,17 @@ int dap_maxlines = 2048;        /* max number of lines in memory file:
                                  */
 #endif
 
+#ifdef DAP_MAXMEM
+int dap_maxmem = DAP_MAXMEM;
+#else
+int dap_maxmem = 1048576; /* memory buffer size for sorting */
+#endif
+#ifdef DAP_TMPDIR
+char *dap_tmpdir = DAP_TMPDIR;
+#else
+char *dap_tmpdir = "dap_tmp";	/* where datasets are stored */
+#endif
+
 /* Memory allocation tracing */
 #ifdef DAP_MEMTRACE
 char *dap_memtrace = DAP_MEMTRACE;
