@@ -2040,6 +2040,7 @@ void input(char varlist[])
     }
 }
 
+/* open an input dataset named fname; if fname is NULL, just close the open input dataset */
 void inset(char *fname)
 {
   int v;
@@ -2059,10 +2060,7 @@ void inset(char *fname)
       dap_out[dap_ono] = (DFILE *) NULL;
     }
   if (!fname)
-    {
-      fputs("(inset) missing input dataset name.\n", dap_err);
-      exit(1);
-    }
+    return;
   if (!(dap_in[dap_ono] = dfopen(fname, "r")))
     {
       fprintf(dap_err, "(inset) can't read data set: %s\n", fname);
