@@ -17,6 +17,9 @@
  *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 
+#include <sys/types.h>
+#include <sys/stat.h>
+#include <fcntl.h>
 #include <unistd.h>
 #include "dap_make.h"
 #include "externs.h"
@@ -99,18 +102,18 @@ static int linediff(char l1[], char l2[])
   return l1[l] != l2[l];
 }
 
+static void dsort(char *origset, char *sortset, int sortvar[], int nfields,
+	   int unique, char *mod, int nmods); /* disk file sorting function */
+
 void sort(char *fname, char *varlist, char *modifiers)
 {
   static int sortinit = 0;	/* memory allocated? */
-  static char *command;
-  int unique;
+  int unique; /* in sorted dataset, include only one of a group of lines with matching keys? */
   int lastun;
   int l;
   int i;
   int v;
   int vn;
-  static char *vname;
-  char sortarg[15];
   char *dsrt0;
   char *dsrt;
   char *dfile;
@@ -128,10 +131,8 @@ void sort(char *fname, char *varlist, char *modifiers)
       startmem = (int *) dap_malloc(sizeof(int) * 2 * dap_maxvar, "");
       start[0] = startmem;
       start[1] = startmem + dap_maxvar;
-      mod = dap_malloc(sizeof(char *) * dap_maxvar, "");
+      mod = dap_malloc(dap_maxvar, "");
       varv = (int *) dap_malloc(sizeof(int) * dap_maxvar, "");
-      command = dap_malloc(dap_linelen + 1, "");
-      vname = dap_malloc(dap_namelen + 1, "");
     }
   if (!fname)
     {
@@ -171,16 +172,16 @@ void sort(char *fname, char *varlist, char *modifiers)
   dap_suffix(dsrt0, fname, ".srt");
   inset(fname);
   outset(dsrt0, "");
+  nvar = dap_list(varlist, varv, dap_maxvar);
+  if (nmods && (nvar != nmods))
+    {
+      fprintf(dap_err,
+	      "(sort) Number of modifiers %d does not equal number of sort variables %d.\n",
+	      nmods, nvar);
+      exit(1);
+    }
   if (fname[0] == '<')
     {
-      nvar = dap_list(varlist, varv, dap_maxvar);
-      if (nmods && (nvar != nmods))
-	{
-	  fprintf(dap_err,
-		  "(sort) Number of modifiers %d does not equal number of sort variables %d.\n",
-		  nmods, nvar);
-	  exit(1);
-	}
       for (nlines = 0, i = dap_ftell(dap_in[0]), newline = 1;
 	   i < dap_in[0]->dfile_ram->rfile_end -
 	     dap_in[0]->dfile_ram->rfile_str; i++)
@@ -222,56 +223,12 @@ void sort(char *fname, char *varlist, char *modifiers)
     }
   else
     {
+      inset(NULL); /* because we had to call inset before and now we don't want it */
       dfile = dap_malloc(strlen(fname) + strlen(dap_setdir) + 2, "");
       dap_name(dfile, fname);
       dsrt = dap_malloc(strlen(dsrt0) + strlen(dap_setdir) + 2, "");
       dap_name(dsrt, dsrt0);
-      sprintf(command, "sed 1d < %s | sort -t '%c'", dfile, SETDELIM);
-      if (unique)
-	strcat(command, " -u");
-      for (l = 0; varlist[l] == ' '; l++)
-	;
-      for (vn = 0; varlist[l]; vn++)
-	{
-	  for (i = 0; varlist[l + i] && varlist[l + i] != ' '; i++)
-	    vname[i] = varlist[l + i];
-	  vname[i] = '\0';
-	  if ((v = dap_varnum(vname)) >= 0)
-	    {
-	      sprintf(sortarg, " -k %d,%d", v + 1, v + 1);
-	      strcat(command, sortarg);
-	      if (nmods)
-		{
-		  if (vn < nmods)
-		    {
-		      if (mod[vn] == 'd')
-			strcat(command, "r");
-		    }
-		  else
-		    {
-		      fputs("(sort) More variables than modifiers.\n", dap_err);
-		      exit(1);
-		    }
-		}
-	    }
-	  else
-	    {
-	      fprintf(dap_err, "(sort) Unknown variable: %s\n",
-		      vname);
-	      exit(1);
-	    }
-	  l += i;
-	  while (varlist[l] == ' ')
-	    l++;
-	}
-      if (vn < nmods)
-	{
-	  fputs("(sort) Fewer variables than modifiers.\n", dap_err);
-	  exit(1);
-	}
-      strcat(command, " >> ");
-      strcat(command, dsrt);
-      system(command);
+      dsort(dfile, dsrt, varv, nvar, unique, mod, nmods);
       dap_free(dsrt, "");
       dap_free(dfile, "");
     }
@@ -2168,3 +2125,535 @@ void join(char *fname, char *partvars, char *valuevar)
   dap_free(varspec, "");
   dap_free(partv, "");
 }
+
+/* dsort -- dap dataset sort HERE
+ *
+ * Usage: sort -f1[d] [-f2[d] ... -fN[d]] [-u] file
+ *
+ * f1 ... fN are field numbers of keys. First field is number 0.
+ * trailing 'd' indicates descending
+ * -u indicates delete lines with duplicate keys
+ */
+
+/* certain variables are inherited from original call to Unix sort,
+ * so must be copied into the variables in this program.
+ */
+/* variables set by fixheader */
+static int *fieldstart; /* starting positions of all the expanded, reordered fields;
+			 * fieldstart[nvars] = position after last field
+			 */
+static int *fieldlen; /* lengths of fields */
+static int *unfield; /* inverse operation of field, set by fixheader */
+static int linelen; /* length of line with fields expanded to max (no delim or terminator) */
+static int keylen; /* length of just the keys */
+static int *keyend; /* ends of the keys */
+static int *sortord; /* 1 for ascending, 0 for descending */
+static int *keymap; /* sortord for each character */
+static int nvars; /* number of variables in dataset */
+
+static char *mem1; /* the memories for reordering, sorting, and writing */
+static char *mem2;
+static int maxlines; /* dap_maxmem / linelen */
+static char **line; /* pointers to the beginnings of lines */
+
+static int *field; /* field numbers of keys */
+static int nfields; /* number of fields specified */
+static int unique; /* -u flag set */
+
+#define TMPLATE "dap_tmp/sortppppppssssss" /* for segment files */
+static char tmplate[] = TMPLATE;
+
+static unsigned long int fpos; /* file position at beginning of segment */
+
+static int strtoi(char *s, int *pi, int *ps)
+{
+  int i;
+
+  for (i = 0; '0' <= *s && *s <= '9'; s++)
+    i = 10 * i + *s - '0';
+  *pi = i;
+  if (*s == 'd')
+    {
+      *ps = 0;
+      s++;
+    }
+  else
+    *ps = 1;
+  return !*s;
+}
+
+static void cleanup(int nseg)
+{
+  int s; /* segment number */
+  char segname[] = TMPLATE;
+
+  for (s = 0; s < nseg; s++)
+    {
+      strcpy(segname, tmplate);
+      sprintf(segname + 18, "%06d", s);
+      unlink(segname);
+    }
+}
+
+static int linecmp(char **s1, char **s2)
+{
+  char *t1, *t2;
+  char *e1;
+
+  /* run through keys until chars differ or at -- not beyond -- end of keys */
+  for (e1 = *s1 + keylen - 1, t1 = *s1, t2 = *s2; t1 < e1 && *t1 == *t2; t1++, t2++)
+    ;
+  return (keymap[t1 - *s1] ? *t1 - *t2 : *t2 - *t1);
+}
+
+/* read in part of orig, sort, and write as segment number s */
+static int sortseg(int orig, int s)
+{
+  char segname[] = TMPLATE;
+  ssize_t nread; /* number of bytes read so far */
+  int seg; /* file descriptor of segment to write */
+  char *m1, *m2; /* for stepping through mem1 and mem2 */
+  int nlines; /* the actual number of lines read */
+  char *lstart1, *lstart2; /* start of line to be read and written */
+  int f; /* index to fields */
+  int last; /* last char gotten from line */
+  int l; /* for stepping through line */
+  char *lp; /* for stepping through fields */
+  char *mp; /* for putting characters back into m */
+  int newfield; /* starting new field in line? */
+  int (*cmp)();
+
+  cmp = &linecmp;
+  nread = read(orig, mem1, dap_maxmem); /* read into mem1, reorder fields into mem2 */
+  if (!nread) /* nothing to read */
+    return 0;
+  strcpy(segname, tmplate);
+  sprintf(segname + 18, "%06d", s);
+  if ((seg = open(segname, O_WRONLY | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR)) < 0)
+    {
+      fprintf(dap_err, "(dsort) can't write %s\n", segname);
+      exit(1);
+    }
+  /* lstart1 is start of line we're getting, but may not get all of */
+  /* need to reorder fields in accordance with fieldstart */
+  for (lstart1 = mem1, lstart2 = mem2, nlines = 0;
+       lstart1 < mem1 + nread && lstart2 + linelen < mem2 + dap_maxmem;
+       lstart1 = m1, lstart2 += linelen)
+    {
+      /* do one line */
+      if (nlines < maxlines)
+	line[nlines++] = lstart2;
+      else
+	break;
+      for (f = 0, m1 = lstart1, last = '\0'; f < nvars; f++) /* for each field */
+	{
+	  for (m2 = lstart2 + fieldstart[unfield[f]];
+	       m1 < mem1 + nread && *m1 != '|' && *m1 != '\n'; )
+	    *m2++ = *m1++;
+	  last = *m1++;
+	  while (m2 < lstart2 + fieldstart[unfield[f] + 1])
+	    *m2++ = '\0';
+	}
+      if (last != '\n') /* didn't finish line */
+	break;
+    }
+  if (last != '\n')
+    --nlines;
+  nread = lstart1 - mem1; /* last line started but not completed */
+  if (!nread)
+    {
+      fputs("(dsort) line longer than buffer\n", dap_err);
+      exit(1);
+    }
+  fpos += nread;
+  qsort(line, nlines, sizeof(char *), cmp);
+  for (l = 0; l < nlines; l++)
+    memcpy(mem1 + l * linelen, line[l], linelen);
+  write(seg, mem1, nlines * linelen);
+  close(seg);
+  return nread;
+}
+
+static char **nextline; /* pointers to next line for each segment */
+
+/* compare the next lines to output of two segments */
+static int nextlinecmp(int *s1, int *s2)
+{
+  return linecmp(nextline + *s1, nextline + *s2);
+}
+
+/* read chunks of seg files into mem1, copy merged into mem2 and write out */
+static void merge(int nseg, int out)
+{
+  int s, s1, s2; /* segment number */
+  int c;
+  int l; /* index to line */
+  int f; /* index to fields */
+  char *segnamemem;
+  char **segname; /* segment names */
+  int segfile;
+  int chunksize; /* chunk of seg file to read at once */
+  int *chunkread; /* amount actually read from segment */
+  char *prevline; /* most recent line produced, for checking uniqueness */
+  unsigned long int *spos; /* position in segment file */
+  int nmore; /* number of segments with more */
+  unsigned int outpos; /* position in mem2 */
+  int *segord; /* order of next lines in segments */
+  int (*scmp)();
+
+  scmp = &nextlinecmp;
+  nextline = (char **) dap_malloc(nseg * sizeof(char *), "nextline");
+  prevline = NULL; /* mark as not yet set */
+  segnamemem = (char *) dap_malloc(nseg * (strlen(TMPLATE) + 1), "segnamemem");
+  segname = (char **) dap_malloc(nseg * sizeof(char *), "segname");
+  spos = (unsigned long int *) dap_malloc(nseg * sizeof(unsigned long int), "spos");
+  segord = (int *) dap_malloc(nseg * sizeof(int), "segord");
+  if (!(chunksize = dap_maxmem / (nseg * linelen) * linelen)) /* read in complete lines */
+    {
+      fputs("(dsort) insufficient memory\n", dap_err);
+      exit(1);
+    }
+  chunkread = (int *) dap_malloc(nseg * sizeof(int), "chunkread");
+  /* set up segnames, nextline, spos */
+  for (s = 0; s < nseg; s++)
+    {
+      segname[s] = segnamemem + s * (strlen(TMPLATE) + 1);
+      strcpy(segname[s], tmplate);
+      sprintf(segname[s] + 18, "%06d", s);
+      nextline[s] = mem1 + s * chunksize;
+      spos[s] = 0; /* at beginning of segment */
+      chunkread[s] = 0; /* nothing yet */
+      segord[s] = s;
+    }
+  /* get initial chunks */
+  for (s = 0; s < nseg; s++)
+    {
+      if ((segfile = open(segname[s], O_RDONLY)) < 0)
+	{
+	  fprintf(dap_err, "(dsort) can't read %s\n", segname[s]);
+	  exit(1);
+	}
+      if ((chunkread[s] = read(segfile, mem1 + s * chunksize, chunksize)) > 0)
+	spos[s] += chunkread[s];
+      else
+	{
+	  fprintf(dap_err, "(dsort) bad initial read of %s\n", segname[s]);
+	  exit(1);
+	}
+      close(segfile);
+    }
+  /* initial sort of next lines to put out: segord[0] is index of that segment */
+  qsort(segord, nseg, sizeof(int), scmp);
+  /* now look for first line to copy to mem2 and move on to next line from that file,
+   * getting new chunk if necessary
+   */
+  for (nmore = nseg, outpos = 0; nmore; )
+    {
+      /* check uniqueness if requested */
+      if (unique)
+	{
+	  if (!prevline)
+	    prevline = (char *) dap_malloc(linelen, "prevline");
+	}
+      /* now need to reformat line */
+      if (!unique || (prevline && linecmp(nextline + segord[0], &prevline)))
+	{
+	  for (f = 0; f < nvars; f++)
+	    {
+	      for (l = fieldstart[unfield[f]]; l < fieldstart[unfield[f] + 1]; l++)
+		{
+		  if (outpos == dap_maxmem) /* if there's no room */
+		    {
+		      write(out, mem2, outpos); /* write it out */
+		      outpos = 0; /* ready for more */
+		    }
+		  if (!(mem2[outpos] = nextline[segord[0]][l]))
+		    break;
+		  outpos++;
+		}
+	      if (outpos == dap_maxmem) /* if there's no room */
+		{
+		  write(out, mem2, outpos); /* write it out */
+		  outpos = 0; /* ready for more */
+		}
+	      if (f < nvars - 1)
+		mem2[outpos++] = '|';
+	      else
+		mem2[outpos++] = '\n';
+	    }
+	}
+      if (unique)
+	{
+	  if (!prevline)
+	    prevline = (char *) dap_malloc(linelen, "prevline");
+	  memcpy(prevline, nextline[segord[0]], linelen);
+	}
+      if ((nextline[segord[0]] += linelen) >=
+	  mem1 + segord[0] * chunksize + chunkread[segord[0]])
+	{ /* if ran out of data in chunk, read the next chunk */
+	  if ((segfile = open(segname[segord[0]], O_RDONLY)) < 0)
+	    {
+	      fprintf(dap_err, "(dsort) can't read %s\n", segname[segord[0]]);
+	      exit(1);
+	    }
+	  if (lseek(segfile, spos[segord[0]], SEEK_CUR) > 0) /* can get to position */
+	    {
+	      if ((chunkread[segord[0]] =
+		   read(segfile, mem1 + segord[0] * chunksize, chunksize)) > 0)
+		{
+		  spos[segord[0]] += chunkread[segord[0]];
+		  nextline[segord[0]] = mem1 + segord[0] * chunksize;
+		}
+	      else
+		--nmore;
+	    }
+	  else
+	    {
+	      chunkread[segord[0]] = 0;
+	      --nmore;
+	    }
+	  close(segfile);
+	}
+      /* now find next line to output: only segment segord[0] has changed, so we just
+       * need to bubble it up as necessary. Number of segments remaining is nmore.
+       */
+      if (chunkread[segord[0]] <= 0) /* seg gone, just have to shift down the others */
+	{
+	  for (s = 0; s < nmore; s++)
+	    segord[s] = segord[s + 1];
+	}
+      else /* bubble up as necessary */
+	{
+	  for (s1 = 1; s1 < nmore; s1++)
+	    {
+	      if (linecmp(nextline + segord[0], nextline + segord[s1]) <= 0)
+		break;
+	    }
+	  s2 = segord[0];
+	  for (s = 0; s < s1 - 1; s++)
+	    segord[s] = segord[s + 1];
+	  segord[s] = s2;
+	}
+    }
+  if (outpos)
+    write(out, mem2, outpos);
+  dap_free(nextline, "");
+  dap_free(prevline, "");
+  dap_free(segnamemem, "");
+  dap_free(segname, "");
+  dap_free(spos, "");
+  dap_free(chunkread, "");
+  dap_free(segord, "");
+}
+
+/* convert code for field into (max) field length */
+static int fieldfix(int code)
+{
+  switch (code)
+    {
+    case DBL:
+      return 12; /* length of coded double */
+    case INT:
+      return 6; /* maximum length of coded int */
+    default: /* length of string */
+      return code;
+    }
+}
+
+/* reorders header for field comparisons and writes out
+ * returns file position after header for resetting pointer
+ * sets values of fieldstart and linelen
+ */
+static int fixheader(char *header, int srt)
+{
+  int headerlen; /* index to header and length of header */
+  int nspaces; /* group variables in header with one internal space */
+  int v; /* index to variables */
+  int gotsign; /* got sign of field spec */
+  int sign; /* sign of field spec */
+  int h; /* index to header */
+  char *h1; /* pointer to header char */
+  int newfield; /* reached end of field, ready for next */
+  int fieldlen1; /* length of one field */
+  int f, f1, f2; /* indexes to fields */
+  int k, k1; /* indexes to keys */
+  char *headerfield[dap_maxvar]; /* pointers to start of field names and specs */
+
+  fieldstart = (int *) dap_malloc((dap_maxvar + 1) * sizeof(int), "fieldstart");
+  fieldstart[0] = 0;
+  fieldlen = (int *) dap_malloc((dap_maxvar + 1) * sizeof(int), "fieldlen");
+  for (headerlen = 0, nspaces = 0, nvars = 0, newfield = 1, linelen = 0;
+       headerlen < dap_linelen && header[headerlen] != '\n';
+       headerlen++)
+    {
+      if (newfield) /* get pointer to start of each field */
+	{
+	  if (nvars)
+	    linelen +=
+	      (fieldlen[nvars - 1] = fieldfix(sign * fieldlen1));  /* no delimiter */
+	  headerfield[nvars++] = header + headerlen;
+	  newfield = 0;
+	}
+      if (header[headerlen] == ' ')
+	{
+	  if (++nspaces == 1)
+	    {
+	      fieldlen1 = 0;
+	      gotsign = 0; /* not set yet */
+	    }
+	  else
+	    {
+	      nspaces = 0;
+	      newfield = 1;
+	    }
+	}
+      else if (nspaces == 1 && header[headerlen] != ' ') /* into length spec */
+	{
+	  if (!gotsign) /* not set yet */
+	    {
+	      if (header[headerlen] == '-')
+		sign = -1;
+	      else
+		{
+		  sign = 1;
+		  gotsign = 1;
+		}
+	    }
+	  if (gotsign)
+	    fieldlen1 = 10 * fieldlen1 + header[headerlen] - '0';
+	  gotsign = 1;
+	}
+    }
+  linelen += (fieldlen[nvars - 1] = fieldfix(sign * fieldlen1)); /* no delimiter */
+  /* make map for sort orders */
+  maxlines = dap_maxmem / linelen;
+  line = (char **) dap_malloc(maxlines * sizeof(char *), "line");
+  if (headerlen == dap_linelen)
+    {
+      header[dap_linelen - 1] = '\0';
+      fprintf(dap_err, "(dsort) header line too long %s\n", header);
+      exit(1);
+    }
+  /* now fill in remainder of field array */
+  for (f = 0, f2 = nfields; f < nvars; f++)
+    {
+      for (f1 = 0; f1 < nfields; f1++)
+	{
+	  if (field[f1] == f)
+	    break;
+	}
+      if (f1 == nfields) /* wasn't there */
+	field[f2++] = f;
+    }
+  /* now invert to create unfield */
+  for (f = 0; f < nvars; f++)
+    unfield[field[f]] = f;
+  /* now set up fieldstarts */
+  for (fieldstart[0] = 0, f = 1; f <= nvars; f++)
+    fieldstart[f] = fieldstart[f - 1] + fieldlen[field[f - 1]];
+  /* now calculate keylen and keyends */
+  keyend = (int *) dap_malloc(nfields * sizeof(int), "keyend");
+  for (f = 0, keylen = 0; f < nfields; f++)
+    {
+      keylen += fieldlen[field[f]];
+      keyend[f] = keylen;
+    }
+  keymap = (int *) dap_malloc(keylen * sizeof(int), "keymap");
+  for (f = 0, k = 0; f < nfields; f++)
+    {
+      for (k1 = 0; k1 < fieldlen[field[f]]; k1++)
+	keymap[k++] = sortord[f];
+    }
+  write(srt, header, headerlen + 1);
+  return headerlen + 1;
+}
+
+static void dsort(char *origset, char *sortset, int sortvar[], int nsort,
+		  int uniq, char *mod, int nmods)
+{
+  int f; /* index to field */
+  int orig; /* orginal file */
+  int nseg; /* number of segments */
+  int srt; /* output file */
+  char *header; /* dap dataset header */
+  int c; /* for skipping and then copying header line */
+  struct stat statbuf; /* for seeing if directory exists */
+
+  header = (char *) dap_malloc(dap_linelen, "header");
+  field = (int *) dap_malloc(dap_maxvar * sizeof(int), "field");
+  unfield = (int *) dap_malloc(dap_maxvar * sizeof(int), "unfield");
+  sortord = (int *) dap_malloc(dap_maxvar * sizeof(int), "dsortord");
+  nfields = nsort;
+  if (!nmods)
+    {
+      for (f = 0; f < nfields; f++)
+	mod[f] = 'i';
+    }
+  else if (nmods != nfields)
+    {
+      fprintf(dap_err, "(dsort) nmods (%d) != nfields (%d)\n", nmods, nfields);
+      exit(1);
+    }
+  unique = uniq;
+  for (f = 0; f < nfields; f++)
+    {
+      sortord[f] = (mod[f] == 'i' ? 1 : 0);
+      field[f] = sortvar[f];
+    }
+  if (stat(dap_tmpdir, &statbuf) < 0)
+    {
+      if (mkdir(dap_tmpdir, (mode_t) 0700) < 0)
+	{
+	  perror(dap_dapname);
+	  exit(1);
+	}
+    }
+  else if (!(statbuf.st_mode & S_IFDIR))
+    {
+      fprintf(dap_err, "%s: non-directory file exists: %s\n", dap_dapname, dap_tmpdir);
+      exit(1);
+    }
+  /* the memories for reordering, sorting, and writing */
+  mem1 = (char *) dap_malloc(2 * dap_maxmem, "mem1");
+  mem2 = mem1 + dap_maxmem;
+  if (!nfields)
+    {
+      fputs("(dsort) no fields specified for sorting\n", dap_err);
+      exit(1);
+    }
+  if ((orig = open(origset, O_RDONLY)) < 0)
+    {
+      fprintf(dap_err, "(dsort) can't read %s\n", origset);
+      exit(1);
+    }
+  if ((srt = open(sortset, O_WRONLY | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR)) < 0)
+    {
+      fprintf(dap_err, "(dsort) can't write %s\n", sortset);
+      exit(1);
+    }
+  /* copy header and reorganize for field comparisons */
+  if (read(orig, header, dap_linelen) < 0)
+    {
+      fputs("(dsort) can't read header\n", dap_err);
+      exit(1);
+    }
+  fpos = fixheader(header, srt);
+  lseek(orig, fpos, SEEK_SET);
+  /* first sort segments of file */
+  sprintf(tmplate + 12, "%06d", getpid() % 1000000); /* for creating segment files */
+  for (nseg = 0; sortseg(orig, nseg); nseg++)
+    lseek(orig, fpos, SEEK_SET);
+  /* then merge */
+  merge(nseg, srt);
+  cleanup(nseg);
+  dap_free(header, "");
+  dap_free(mem1, "");
+  dap_free(fieldstart, "");
+  dap_free(line, "");
+  dap_free(field, "");
+  dap_free(unfield, "");
+  dap_free(fieldlen, "");
+  dap_free(sortord, "");
+  dap_free(keyend, "");
+  dap_free(keymap, "");
+}
