@@ -1,6 +1,6 @@
 /* dap3.c -- plotting routines */
 
-/*  Copyright (C) 2001, 2002, 2003 Free Software Foundation, Inc.
+/*  Copyright (C) 2001, 2002, 2003, 2005 Free Software Foundation, Inc.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -388,8 +388,9 @@ static int dblcmp(double *x, double *y)
 #define INVSQRTPI 0.56418958354775628
 #define TWOOSQRTPI 1.12837916709551257
 #define SQRTPI 1.77245385090551602729
+#define SIXOVERPI 1.90985931710274381
 
-static double comb(int n, int k)
+static double comb(int n, int k) /* n combination k */
 {
   double dn, dk;
   double c;
@@ -403,6 +404,12 @@ static double comb(int n, int k)
 static double dnmk;
 static double dkm1;
 
+/* To be able to integrate on a finite range -1 to 1,
+ * we obtain a normal variate x as t / sqrt(1 - t * t), where -1 < t < 1.
+ * Then orderf(t) is x times the prob density of the order statistic of x,
+ * except for the combinatorial coefficient, but in terms of t, so that
+ * we can integrate to find the expected value.
+ */
 static double orderf(double t)
 {
   double x;
@@ -416,25 +423,48 @@ static double orderf(double t)
   if (finite(tmp2))
     return exp(tmp2) * t / (tmp1 * tmp1);
   return 0.0;
-  /* Original code:
-     return exp(dkm1 * log(probz(x)) + dnmk * log(probz(-x)) - 0.5 * x * x) *
-     t / (tmp1 * tmp1);
-  */
 }
 
-#define NSTEPS 128
+#define NSTEPS 512
 
-static void geta(double a[], int n)
+static void geta(double a[], int n) /* get (negative of) Shaprio-Wilk "a" array */
 {
-  int k;
-  double dn;
+  int k; /* index to "a" array */
+  double dn; /* number of points, as double */
   double c;
 
-  for (k = 0; k < n / 2; k++)
+  if (n < 7) /* get "exact" values for "a" array */
     {
+      switch (n)
+	{
+	case 3:
+	  a[0] = 0.707107;
+	  break;
+	case 4:
+	  a[0] = 0.687155;
+	  a[1] = 0.166787;
+	  break;
+	case 5:
+	  a[0] = 0.664647;
+	  a[1] = 0.241337;
+	  break;
+	case 6:
+	  a[0] = 0.643105;
+	  a[1] = 0.280635;
+	  a[2] = 0.0875196;
+	  break;
+	}
+      return;
+    }
+  /* Here we use the approximations on page 117 of
+   * Royston, Appl. Statist. (1982), 31, No. 2, pp. 115-124
+   */
+  for (k = 0; k < n / 2; k++)
+    { /* first get the expected values of the order statistics of n normal variates */
       dnmk = (double) (n - k - 1);
       dkm1 = (double) k;
-      a[k] = ((double) k + 1) * comb(n, k + 1) * INVSQ2PI * dap_simp(&orderf, -1.0, 1.0, NSTEPS);
+      a[k] = ((double) k + 1) * comb(n, k + 1) * INVSQ2PI *
+	dap_simp(&orderf, -1.0, 1.0, NSTEPS);
     }
   dn = (double) n;
   if (n <= 20)
@@ -463,119 +493,104 @@ static void geta(double a[], int n)
     a[k] *= c;
 }
 
-#define NSIMUL 10000
+/* Approximate the probability that W <= w0:
+ * For n = 3: direct calculation;
+ * For 4 <= n <= 6: transformation to n = 3, then direct calculation;
+ * For 7 <= n: normalizing transformation
+ */
 
-typedef struct swvalstr
+static double clambda1[3] =
+{ 0.118898, 0.133414, 0.327907 };
+static double clambda2[6] =
+{ 0.480385, 0.318828, 0.0, -0.0241665, 0.00879701, 0.002989646 };
+static double clogmu1[4] =
+{ -0.37542, -0.492145, -1.124332, -0.199422 };
+static double clogmu2[6] =
+{ -1.91487, -1.37888, -0.04183209, 0.1066339, -0.03513666, -0.01504614 };
+static double clogsigma1[4] =
+{ -3.15805, 0.729399, 3.01855, 1.558776 };
+static double clogsigma2[7] =
+{ -3.73538, -1.015807, -0.331885, 0.1773538, -0.01638782, -0.03215018, 0.003852646 };
+
+static double au[3][2][5] =
+{ { { -1.26233, 1.87969, 0.0649583, -0.0475604, -0.0139682 },
+    { -0.287696, 1.78953, -0.180114, 0.0, 0.0 } },
+  { { -2.28135, 2.26186, 0.0, 0.0, -0.00865763 },
+    { -1.63638, 5.60924, -3.63738, 1.08439, 0.0 } },
+  { { -3.30623, 2.76287, -0.83484, 1.20857, -0.507590 },
+    { -5.991908, 21.04575, -24.58061, 13.78661, -2.835295 } }
+};
+
+static double lowhigh[3][2] =
+{ { -3.8, 8.6 }, { -3.0, 5.8 }, { -4.0, 5.4 } };
+
+/* evaluate a polynomial of degree n */
+static double poly(double c[], double x, int n)
 {
-  double swd;
-  struct swvalstr *swleft;
-  struct swvalstr *swright;
-  struct swvalstr *swup;
-} swval;
+  double p;
 
+  for (p = c[n]; --n >= 0; )
+    p = p * x + c[n];
+  return p;
+}
 
-static double probw(double swa[], int n, double w0)
+static double probw(int n, double w0, double a1)
 {
-  int s;
-  int small;
-  swval *y;
-  double yval;
-  swval *py;
-  swval *ppy;
-  int ny;
-  double dny;
-  double sum, ss;
-  double vtmp, tmp;
-  double w;
-  int (*cmp)();
+  double u; /* transformed from w0 */
+  double d, lambda, logmu, logsigma;
+  double y;
+  double eps;
+  int r; /* range index */
 
-  cmp = &dblcmp;
-  y = (swval *) dap_malloc(n * sizeof(swval), "");
-  for (s = 0, small = 0; s < NSIMUL; s++)
+  if (n == 3)
+    return SIXOVERPI * (asin(sqrt(w0)) - asin(sqrt(0.75)));
+  if (n >= 7)
     {
-      ny = 0;
-      y[ny].swd = varnorm();
-      y[ny].swleft = NULL;
-      y[ny].swright = NULL;
-      y[ny].swup = NULL;
-      for (ny++; ny < n; ny++)
+      if (n <= 20)
+	d = 3.0;
+      else
+	d = 5.0;
+      d = log((double) n) - d;
+      if (n <= 20)
 	{
-	  yval = varnorm();
-	  y[ny].swd = yval;
-	  y[ny].swleft = NULL;
-	  y[ny].swright = NULL;
-	  for (py = &y[0]; ; py = ppy)
-	    {
-	      if (yval >= py->swd)
-		{
-		  ppy = py->swright;
-		  if (!ppy)
-		    {
-		      py->swright = &y[ny];
-		      y[ny].swup = py;
-		      break;
-		    }
-		}
-	      else
-		{
-		  ppy = py->swleft;
-		  if (!ppy)
-		    {
-		      py->swleft = &y[ny];
-		      y[ny].swup = py;
-		      break;
-		    }
-		}
-	    }
+	  lambda = poly(clambda1, d, 2);
+	  logmu = poly(clogmu1, d, 3);
+	  logsigma = poly(clogsigma1, d, 3);
 	}
-      for (py = &y[0]; (ppy = py->swleft); py = ppy)
-	;
-      for (ny = 0, sum = 0.0, ss = 0.0, w = 0.0; py; )
+      else
 	{
-	  dny = (double) ny;
-	  vtmp = py->swd;
-	  if (ny < n / 2)
-	    w += swa[ny] * vtmp;
-	  else if (ny > n / 2)
-	    w -= swa[n - 1 - ny] * vtmp;
-	  py->swd = 0.0 / 0.0;
-	  if (ny)
-	    {
-	      tmp = sum - dny * vtmp;
-	      ss += tmp * tmp / (dny * (dny + 1.0));
-	    }
-	  sum += vtmp;
-	  ny++;
-	  if ((ppy = py->swright))
-	    {
-	      for (py = ppy; (ppy = py->swleft); py = ppy)
-		;
-	    }
-	  else if ((ppy = py->swup))
-	    {
-	      for (py = ppy; py && !finite(py->swd); py = py->swup)
-		;
-	    }
-	  else
-	    break;
+	  lambda = poly(clambda2, d, 5);
+	  logmu = poly(clogmu2, d, 5);
+	  logsigma = poly(clogsigma2, d, 6);
 	}
-      if (ny < n)
+      y = pow(1.0 - w0, lambda);
+      return probz(-(y - exp(logmu)) / exp(logsigma));
+    }
+  else
+    {
+      if (w0 >= 1.0)
+	return 1.0;
+      eps = a1 * a1 * (1.0 + 1.0 / ((double) (n - 1)));
+      if (w0 <= eps)
+	return 0.0;
+      u = log((w0 - eps) / (1.0 - w0));
+      if (w0 < lowhigh[n - 4][0] || w0 > lowhigh[n - 4][1])
+	return 0.0 / 0.0;
+      if (u < 1.4)
+	r = 0;
+      else
 	{
-	  fprintf(dap_err,
-		  "(probw) Only collected %d values from tree of %d values.\n",
-		  ny, n);
-	  exit(1);
+	  r = 1;
+	  u = log(u);
+	  if (!finite(u))
+	    return 0.0 / 0.0;
 	}
-      /*
-	for (w = 0.0, ny = 0; ny < n / 2; ny++)
-	w += swa[ny] * (y[ny] - y[n - 1 - ny]);
-      */
-      w *= w / ss;
-      if (w < w0)
-	small++;
+      u = poly(au[n - 4][r], u, 4);
+      if (r)
+	u = exp(u);
+      u = exp(u);
+      return SIXOVERPI * (asin(sqrt((u + 0.75) / (1.0 + u))) - asin(sqrt(0.75)));
     }
-  dap_free(y, "");
-  return ((double) small) / ((double) NSIMUL);
 }
 
 /* Display, if requested, q-q plot, and perform the Shapiro-Wilk
@@ -585,36 +600,39 @@ static void normal1(double x[], double y[], int nobs,
 		    char varname[], char varlabel[], int markv[], int nmark,
 		    pict *p, pict *l, pict *a, int pn)
 {
-  int titlelen;
-  int r;
-  double dr;
-  double dnobsp25;
-  double sum;
-  double ss;
-  double vtmp;
-  double tmp;
-  double sd;
-  double minx, maxx;
-  double *swa;
-  int k;
-  double w;
-  double prob;
+  int titlelen; /* length of title string, +1, for caption */
+  int r; /* index to sorted values (of y), i.e., rank */
+  double dr; /* r converted to double */
+  double dnobsp25; /* number of obs + 1/4, as double, for denom for computing x */
+  double sum; /* for incremental computation of corrected sum of squares */
+  double ss; /* corrected sum of squares */
+  double vtmp; /* temp for incremental computation of corrected sum of squares */
+  double tmp; /* temp for incremental computation of corrected sum of squares */
+  double sd; /* standard deviation: needed for slope of line in q-q plot */
+  double minx, maxx; /* min and max values of x, for q-q plot */
+  double *swa; /* array of Shapiro-Wilk "a" coefficients */
+  int k; /* index of "a" array */
+  double w; /* Shapiro-Wilk statistic */
+  double prob; /* probability that W <= computed Shapiro-Wilk statistic */
   int v;
   char *caption;
-  int (*cmp)();
+  int (*cmp)(); /* address of function for comparing doubles through pointers */
 
-  dap_swap();
+  dap_swap(); /* get back to part of file we want to work on */
   cmp = &dblcmp;
   if (dap_title)
     titlelen = strlen(dap_title) + 1;
   else
     titlelen = 0;
-  if (p)	/* if q-q plot requested */
+  if (p) /* if q-q plot requested, we have pointer to linked list of picts
+	  * and we need to allocate caption string: 47??
+          */
     caption = dap_malloc(dap_linelen + titlelen + 47, "dap_linelen");
   qsort((void *) y, (size_t) nobs, (size_t) sizeof(double), cmp);
+  /* q-q plot has denominator n + 1/4, numerator rank - 3/8 */
   dnobsp25 = ((double) nobs) + 0.25;
   for (r = 0, sum = 0.0, ss = 0.0, minx = 0.0, maxx = 0.0; r < nobs; r++)
-    {
+    { /* for each y, compute an x for plotting: inv std norm(numerator/denominator) */
       dr = (double) r;
       x[r] = -zpoint((dr + 0.625) / dnobsp25);
       if (x[r] < minx)
@@ -623,33 +641,34 @@ static void normal1(double x[], double y[], int nobs,
 	maxx = x[r];
       vtmp = y[r];
       if (r)
-	{
+	{ /* incremental method for computing corrected sum of squares */
 	  tmp = sum - dr * vtmp;
 	  ss += tmp * tmp / (dr * (dr + 1.0));
 	}
       sum += vtmp;
     }
-  sd = sqrt(ss / ((double) (nobs - 1)));
+  sd = sqrt(ss / ((double) (nobs - 1))); /* need sd to plot line for q-q plot */
   if (sd == 0.0)
     {
       fprintf(dap_err, "(normal1) Zero standard deviation for %s\n", varname);
       exit(1);
     }
-  if (6 < nobs && nobs <= 500)
+  if (3 <= nobs && nobs <= 2000)
     {
       swa = (double *) dap_malloc(nobs / 2 * sizeof(double), "");
-      geta(swa, nobs);
-      for (w = 0.0, k = 0; k < nobs / 2; k++)
+      geta(swa, nobs); /* compute Shapiro-Wilk "a" coefficients */
+      for (w = 0.0, k = 0; k < nobs / 2; k++) /* compute W */
 	w += swa[k] * (y[k] - y[nobs - 1 - k]);
       w *= w / ss;
-      if ((prob = probw(swa, nobs, w)) < 0.001)
+      if ((prob = probw(nobs, w, swa[0])) < 0.001)
 	prob = 0.001;
       dap_free(swa, "");
       if (p)	/* if q-q plot requested */
 	sprintf(caption, "q-q plot: W|0| = %.4f, P[W < W|0|] = %.3f", w, prob);
       dap_head(markv, nmark);
       fprintf(dap_lst,
-	      "Shapiro-Wilk test:\nW0 = %.4f, P[W < W0] = %.3f\n", w, prob);
+	      "Shapiro-Wilk test for %s:\nW0 = %.4f, P[W < W0] = %.3f\n",
+	      varname, w, prob);
     }
   else if (p)	/* if q-q plot requested */
     strcpy(caption, "q-q plot");
@@ -685,7 +704,7 @@ static void normal1(double x[], double y[], int nobs,
       strcpy(l[pn].pict_type, "LINE");
       pict_line(l + pn, minx, sd * minx + sum, maxx, sd * maxx + sum);
       p[pn].pict_r = 0.05 *
-	pict_autoaxes(p + pn, "z", varlabel, "00", NULL, NULL, caption, PORTRAIT);
+	pict_autoaxes(p + pn, "z", varlabel, "-0 NXDIGITS3", NULL, NULL, caption, PORTRAIT);
       free(caption);
     }
   dap_swap();
